@@ -76,6 +76,9 @@ type NodeDeleteStatus struct {
 	nodeDeleteResults map[string]error
 }
 
+// Get current time. Proxy for unit tests.
+var now func() time.Time = time.Now
+
 // IsDeleteInProgress returns true if a node is being deleted.
 func (n *NodeDeleteStatus) IsDeleteInProgress() bool {
 	n.Lock()
@@ -355,7 +358,7 @@ func (sd *ScaleDown) UpdateUnneededNodes(
 
 	currentlyUnneededNodes := make([]*apiv1.Node, 0)
 	// Only scheduled non expendable pods and pods waiting for lower priority pods preemption can prevent node delete.
-	nonExpendablePods := FilterOutExpendablePods(pods, sd.context.ExpendablePodsPriorityCutoff)
+	nonExpendablePods := filterOutExpendablePods(pods, sd.context.ExpendablePodsPriorityCutoff)
 	nodeNameToNodeInfo := scheduler_util.CreateNodeNameToInfoMap(nonExpendablePods, nodes)
 	utilizationMap := make(map[string]simulator.UtilizationInfo)
 
@@ -572,6 +575,50 @@ func (sd *ScaleDown) mapNodesToStatusScaleDownNodes(nodes []*apiv1.Node, nodeGro
 	return result
 }
 
+// SoftTaintUnneededNodes manage soft taints of unneeded nodes.
+func (sd *ScaleDown) SoftTaintUnneededNodes(allNodes []*apiv1.Node) (errors []error) {
+	defer metrics.UpdateDurationFromStart(metrics.ScaleDownSoftTaintUnneeded, time.Now())
+	apiCallBudget := sd.context.AutoscalingOptions.MaxBulkSoftTaintCount
+	timeBudget := sd.context.AutoscalingOptions.MaxBulkSoftTaintTime
+	skippedNodes := 0
+	startTime := now()
+	for _, node := range allNodes {
+		if deletetaint.HasToBeDeletedTaint(node) {
+			// Do not consider nodes that are scheduled to be deleted
+			continue
+		}
+		alreadyTainted := deletetaint.HasDeletionCandidateTaint(node)
+		_, unneeded := sd.unneededNodes[node.Name]
+
+		// Check if expected taints match existing taints
+		if unneeded != alreadyTainted {
+			if apiCallBudget <= 0 || now().Sub(startTime) >= timeBudget {
+				skippedNodes++
+				continue
+			}
+			apiCallBudget--
+			if unneeded && !alreadyTainted {
+				err := deletetaint.MarkDeletionCandidate(node, sd.context.ClientSet)
+				if err != nil {
+					errors = append(errors, err)
+					klog.Warningf("Soft taint on %s adding error %v", node.Name, err)
+				}
+			}
+			if !unneeded && alreadyTainted {
+				_, err := deletetaint.CleanDeletionCandidate(node, sd.context.ClientSet)
+				if err != nil {
+					errors = append(errors, err)
+					klog.Warningf("Soft taint on %s removal error %v", node.Name, err)
+				}
+			}
+		}
+	}
+	if skippedNodes > 0 {
+		klog.V(4).Infof("Skipped adding/removing soft taints on %v nodes - API call limit exceeded", skippedNodes)
+	}
+	return
+}
+
 // TryToScaleDown tries to scale down the cluster. It returns a result inside a ScaleDownStatus indicating if any node was
 // removed and error if such occurred.
 func (sd *ScaleDown) TryToScaleDown(allNodes []*apiv1.Node, pods []*apiv1.Pod, pdbs []*policyv1.PodDisruptionBudget, currentTime time.Time) (*status.ScaleDownStatus, errors.AutoscalerError) {
@@ -682,9 +729,9 @@ func (sd *ScaleDown) TryToScaleDown(allNodes []*apiv1.Node, pods []*apiv1.Pod, p
 
 	findNodesToRemoveStart := time.Now()
 	// Only scheduled non expendable pods are taken into account and have to be moved.
-	nonExpendablePods := FilterOutExpendablePods(pods, sd.context.ExpendablePodsPriorityCutoff)
+	nonExpendablePods := filterOutExpendablePods(pods, sd.context.ExpendablePodsPriorityCutoff)
 	// We look for only 1 node so new hints may be incomplete.
-	nodesToRemove, _, _, err := simulator.FindNodesToRemove(candidates, nodesWithoutMaster, nonExpendablePods, sd.context.ClientSet,
+	nodesToRemove, _, _, err := simulator.FindNodesToRemove(candidates, nodesWithoutMaster, nonExpendablePods, sd.context.ListerRegistry,
 		sd.context.PredicateChecker, 1, false,
 		sd.podLocationHints, sd.usageTracker, time.Now(), pdbs)
 	findNodesToRemoveDuration = time.Now().Sub(findNodesToRemoveStart)
@@ -800,7 +847,7 @@ func getEmptyNodes(candidates []*apiv1.Node, pods []*apiv1.Pod, maxEmptyBulkDele
 			if checkResult.exceeded {
 				continue
 			}
-			available -= 1
+			available--
 			availabilityMap[nodeGroup.Id()] = available
 			result = append(result, node)
 		}
@@ -947,7 +994,7 @@ func evictPod(podToEvict *apiv1.Pod, client kube_client.Interface, recorder kube
 	}
 	klog.Errorf("Failed to evict pod %s, error: %v", podToEvict.Name, lastError)
 	recorder.Eventf(podToEvict, apiv1.EventTypeWarning, "ScaleDownFailed", "failed to delete pod for ScaleDown")
-	return fmt.Errorf("Failed to evict pod %s/%s within allowed timeout (last error: %v)", podToEvict.Namespace, podToEvict.Name, lastError)
+	return fmt.Errorf("failed to evict pod %s/%s within allowed timeout (last error: %v)", podToEvict.Namespace, podToEvict.Name, lastError)
 }
 
 // Performs drain logic on the node. Marks the node as unschedulable and later removes all pods, giving
@@ -1011,21 +1058,6 @@ func drainNode(node *apiv1.Node, pods []*apiv1.Pod, client kube_client.Interface
 		errors.TransientError, "Failed to drain node %s/%s: pods remaining after timeout", node.Namespace, node.Name)
 }
 
-// cleanToBeDeleted cleans ToBeDeleted taints.
-func cleanToBeDeleted(nodes []*apiv1.Node, client kube_client.Interface, recorder kube_record.EventRecorder) {
-	for _, node := range nodes {
-		cleaned, err := deletetaint.CleanToBeDeleted(node, client)
-		if err != nil {
-			klog.Warningf("Error while releasing taints on node %v: %v", node.Name, err)
-			recorder.Eventf(node, apiv1.EventTypeWarning, "ClusterAutoscalerCleanup",
-				"failed to clean toBeDeletedTaint: %v", err)
-		} else if cleaned {
-			klog.V(1).Infof("Successfully released toBeDeletedTaint on node %v", node.Name)
-			recorder.Eventf(node, apiv1.EventTypeNormal, "ClusterAutoscalerCleanup", "marking the node as schedulable")
-		}
-	}
-}
-
 // Removes the given node from cloud provider. No extra pre-deletion actions are executed on
 // the Kubernetes side.
 func deleteNodeFromCloudProvider(node *apiv1.Node, cloudProvider cloudprovider.CloudProvider,
